"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var LinkResolverPlugin_1;
Object.defineProperty(exports, "__esModule", { value: true });
exports.LinkResolverPlugin = void 0;
const components_1 = require("../components");
const converter_events_1 = require("../converter-events");
const models_1 = require("../../models");
const declarationReference_1 = require("../comments/declarationReference");
const ts = require("typescript");
const assert_1 = require("assert");
const utils_1 = require("../../utils");
const urlPrefix = /^(http|ftp)s?:\/\//;
const brackets = /\[\[([^\]]+)\]\]/g;
/**
 * A plugin that resolves `{@link Foo}` tags.
 */
let LinkResolverPlugin = LinkResolverPlugin_1 = class LinkResolverPlugin extends components_1.ConverterComponent {
    /**
     * Create a new LinkResolverPlugin instance.
     */
    initialize() {
        super.initialize();
        this.owner.on(converter_events_1.ConverterEvents.RESOLVE_END, this.onResolve, this, -300);
    }
    /**
     * Find all old style double bracket references to symbols within the given text and transform them into a link.
     *
     * @param text  The text that should be parsed.
     * @returns The text with symbol references replaced by links.
     */
    replaceBrackets(reflection, text) {
        const parts = [];
        let begin = 0;
        brackets.lastIndex = 0;
        for (const match of text.matchAll(brackets)) {
            if (begin != match.index) {
                parts.push({
                    kind: "text",
                    text: text.substring(begin, match.index),
                });
            }
            begin = match.index + match[0].length;
            const content = match[1];
            const { target, caption } = LinkResolverPlugin_1.splitLinkText(content);
            if (urlPrefix.test(target)) {
                parts.push({
                    kind: "inline-tag",
                    tag: "@link",
                    text: caption,
                    target,
                });
            }
            else {
                const targetRefl = reflection.findReflectionByName(target);
                if (targetRefl) {
                    parts.push({
                        kind: "inline-tag",
                        tag: "@link",
                        text: caption,
                        target: targetRefl,
                    });
                }
                else {
                    if (this.validation.invalidLink) {
                        this.application.logger.warn("Failed to find target: " + content);
                    }
                    parts.push({
                        kind: "inline-tag",
                        tag: "@link",
                        text: content,
                    });
                }
            }
        }
        parts.push({
            kind: "text",
            text: text.substring(begin),
        });
        return parts;
    }
    onResolve(context) {
        for (const reflection of Object.values(context.project.reflections)) {
            this.processReflection(reflection);
        }
        let warned = false;
        const warn = () => {
            if (!warned) {
                warned = true;
                this.application.logger.warn(`README: Comment [[target]] style links are deprecated and will be removed in 0.24`);
            }
        };
        if (context.project.readme) {
            context.project.readme = this.processParts(context.project, context.project.readme, warn);
        }
    }
    processReflection(reflection) {
        const comment = reflection.comment;
        if (!comment)
            return;
        let warned = false;
        const warn = () => {
            if (!warned) {
                warned = true;
                this.application.logger.warn(`${reflection.getFriendlyFullName()}: Comment [[target]] style links are deprecated and will be removed in 0.24`);
            }
        };
        comment.summary = this.processParts(reflection, comment.summary, warn);
        for (const tag of comment.blockTags) {
            tag.content = this.processParts(reflection, tag.content, warn);
        }
    }
    processParts(reflection, parts, warn) {
        return parts.flatMap((part) => this.processPart(reflection, part, warn));
    }
    processPart(reflection, part, warn) {
        if (part.kind === "text" && brackets.test(part.text)) {
            warn();
            return this.replaceBrackets(reflection, part.text);
        }
        if (part.kind === "inline-tag") {
            if (part.tag === "@link" ||
                part.tag === "@linkcode" ||
                part.tag === "@linkplain") {
                return resolveLinkTag(reflection, part, (msg) => {
                    if (this.validation.invalidLink) {
                        this.application.logger.warn(msg);
                    }
                });
            }
        }
        return part;
    }
    /**
     * Split the given link into text and target at first pipe or space.
     *
     * @param text  The source string that should be checked for a split character.
     * @returns An object containing the link text and target.
     */
    static splitLinkText(text) {
        let splitIndex = text.indexOf("|");
        if (splitIndex === -1) {
            splitIndex = text.search(/\s/);
        }
        if (splitIndex !== -1) {
            return {
                caption: text
                    .substring(splitIndex + 1)
                    .replace(/\n+/, " ")
                    .trim(),
                target: text.substring(0, splitIndex).trim(),
            };
        }
        else {
            return {
                caption: text,
                target: text,
            };
        }
    }
};
__decorate([
    (0, utils_1.BindOption)("validation")
], LinkResolverPlugin.prototype, "validation", void 0);
LinkResolverPlugin = LinkResolverPlugin_1 = __decorate([
    (0, components_1.Component)({ name: "link-resolver" })
], LinkResolverPlugin);
exports.LinkResolverPlugin = LinkResolverPlugin;
function resolveLinkTag(reflection, part, warn) {
    let pos = 0;
    const end = part.text.length;
    // Try to parse one
    const declRef = (0, declarationReference_1.parseDeclarationReference)(part.text, pos, end);
    let target;
    if (declRef) {
        // Got one, great! Try to resolve the link
        target = resolveDeclarationReference(reflection, declRef[0]);
        pos = declRef[1];
    }
    // If resolution via a declaration reference failed, revert to the legacy "split and check"
    // method... this should go away in 0.24, once people have had a chance to migrate any failing links.
    if (!target) {
        const resolved = legacyResolveLinkTag(reflection, part);
        if (resolved) {
            warn(`Failed to resolve {@link ${part.text}} in ${reflection.getFriendlyFullName()} with declaration references. This link will break in v0.24.`);
        }
        return resolved;
    }
    // Remaining text after an optional pipe is the link text, so advance
    // until that's consumed.
    while (pos < end && ts.isWhiteSpaceLike(part.text.charCodeAt(pos))) {
        pos++;
    }
    if (pos < end && part.text[pos] === "|") {
        pos++;
    }
    part.target = target;
    part.text = part.text.substring(pos).trim() || target.name;
    return part;
}
function legacyResolveLinkTag(reflection, part) {
    const { caption, target } = LinkResolverPlugin.splitLinkText(part.text);
    if (urlPrefix.test(target)) {
        part.text = caption;
        part.target = target;
    }
    else {
        const targetRefl = reflection.findReflectionByName(target);
        if (targetRefl) {
            part.text = caption;
            part.target = targetRefl;
        }
    }
    return part;
}
function resolveDeclarationReference(reflection, ref) {
    let high = [];
    let low = [];
    if (ref.moduleSource) {
        high =
            reflection.project.children?.filter((c) => c.kindOf(models_1.ReflectionKind.SomeModule) &&
                c.name === ref.moduleSource) || [];
    }
    else if (ref.resolutionStart === "global") {
        high.push(reflection.project);
    }
    else {
        (0, assert_1.ok)(ref.resolutionStart === "local");
        // TypeScript's behavior is to first try to resolve links via variable scope, and then
        // if the link still hasn't been found, check either siblings (if comment belongs to a member)
        // or otherwise children.
        let refl = reflection;
        if (refl.kindOf(models_1.ReflectionKind.ExportContainer)) {
            high.push(refl);
        }
        while (refl.parent) {
            refl = refl.parent;
            if (refl.kindOf(models_1.ReflectionKind.ExportContainer)) {
                high.push(refl);
            }
        }
        if (reflection.kindOf(models_1.ReflectionKind.SomeMember)) {
            high.push(reflection.parent);
        }
        else if (reflection.kindOf(models_1.ReflectionKind.SomeSignature) &&
            reflection.parent.kindOf(models_1.ReflectionKind.SomeMember)) {
            high.push(reflection.parent.parent);
        }
        else if (high[0] !== reflection) {
            high.push(reflection);
        }
    }
    if (ref.symbolReference) {
        for (const part of ref.symbolReference.path || []) {
            const high2 = high;
            high = [];
            const low2 = low;
            low = [];
            for (const refl of high2) {
                const resolved = resolveSymbolReferencePart(refl, part);
                high.push(...resolved.high);
                low.push(...resolved.low);
            }
            for (const refl of low2) {
                const resolved = resolveSymbolReferencePart(refl, part);
                low.push(...resolved.high);
                low.push(...resolved.low);
            }
        }
        if (ref.symbolReference.meaning) {
            high = filterMapByMeaning(high, ref.symbolReference.meaning);
            low = filterMapByMeaning(low, ref.symbolReference.meaning);
        }
    }
    return high[0] || low[0];
}
function filterMapByMeaning(reflections, meaning) {
    return (0, utils_1.filterMap)(reflections, (refl) => {
        const kwResolved = resolveKeyword(refl, meaning.keyword) || [];
        return kwResolved[meaning.index || 0];
    });
}
function resolveKeyword(refl, kw) {
    switch (kw) {
        case undefined:
            return refl instanceof models_1.DeclarationReflection && refl.signatures
                ? refl.signatures
                : [refl];
        case "class":
            if (refl.kindOf(models_1.ReflectionKind.Class))
                return [refl];
            break;
        case "interface":
            if (refl.kindOf(models_1.ReflectionKind.Interface))
                return [refl];
            break;
        case "type":
            if (refl.kindOf(models_1.ReflectionKind.SomeType))
                return [refl];
            break;
        case "enum":
            if (refl.kindOf(models_1.ReflectionKind.Enum))
                return [refl];
            break;
        case "namespace":
            if (refl.kindOf(models_1.ReflectionKind.SomeModule))
                return [refl];
            break;
        case "function":
            if (refl.kindOf(models_1.ReflectionKind.FunctionOrMethod)) {
                return refl.signatures;
            }
            break;
        case "var":
            if (refl.kindOf(models_1.ReflectionKind.Variable))
                return [refl];
            break;
        case "new":
        case "constructor":
            if (refl.kindOf(models_1.ReflectionKind.ClassOrInterface | models_1.ReflectionKind.TypeLiteral)) {
                const ctor = refl.children?.find((c) => c.kindOf(models_1.ReflectionKind.Constructor));
                return ctor?.signatures;
            }
            break;
        case "member":
            if (refl.kindOf(models_1.ReflectionKind.SomeMember))
                return [refl];
            break;
        case "event":
            // Never resolve. Nobody should use this.
            // It's required by the grammar, but is not documented by TypeDoc
            // nor by the comments in the grammar.
            break;
        case "call":
            return refl.signatures;
        case "index":
            if (refl.indexSignature) {
                return [refl.indexSignature];
            }
            break;
        case "complex":
            if (refl.kindOf(models_1.ReflectionKind.SomeType))
                return [refl];
            break;
    }
}
function resolveSymbolReferencePart(refl, path) {
    let high = [];
    let low = [];
    if (!(refl instanceof models_1.ContainerReflection) || !refl.children) {
        return { high, low };
    }
    switch (path.navigation) {
        // Grammar says resolve via "exports"... as always, reality is more complicated.
        // Check exports first, but also allow this as a general purpose "some child" operator
        // so that resolution doesn't behave very poorly with projects using JSDoc style resolution.
        // Also is more consistent with how TypeScript resolves link tags.
        case ".":
            high = refl.children.filter((r) => r.name === path.path &&
                (r.kindOf(models_1.ReflectionKind.SomeExport) || r.flags.isStatic));
            low = refl.children.filter((r) => r.name === path.path &&
                (!r.kindOf(models_1.ReflectionKind.SomeExport) || !r.flags.isStatic));
            break;
        // Resolve via "members", interface children, class instance properties/accessors/methods,
        // enum members, type literal properties
        case "#":
            high = refl.children.filter((r) => {
                return (r.name === path.path &&
                    r.kindOf(models_1.ReflectionKind.SomeMember) &&
                    !r.flags.isStatic);
            });
            break;
        // Resolve via "locals"... treat this as a stricter `.` which only supports traversing
        // module/namespace exports since TypeDoc doesn't support documenting locals.
        case "~":
            if (refl.kindOf(models_1.ReflectionKind.SomeModule | models_1.ReflectionKind.Project)) {
                high = refl.children.filter((r) => r.name === path.path);
            }
            break;
    }
    return { high, low };
}
